"""
Tests for relationship module
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

import unittest
import numpy as np
import scipy.stats
from uncertainties import unumpy as unp
from numpy.testing import assert_almost_equal, assert_equal
from uravu import utils
from uravu.relationship import Relationship
from uravu.distribution import Distribution
from uravu.axis import Axis


TEST_Y = []
for i in np.arange(1, 9, 1):
    TEST_Y.append(Distribution(scipy.stats.norm.rvs(loc=i, scale=0.5, size=200, random_state=np.random.RandomState(1))))
TEST_X = np.arange(1, 9, 1)


class TestRelationship(unittest.TestCase):
    """
    Tests for the relationship module and class.
    """
    def test_function_init(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y)
        assert_equal(r.function, utils.straight_line)

    def test_abscissa_init(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y)
        assert_equal(isinstance(r.abscissa, np.ndarray), True)
        assert_equal(r.abscissa, TEST_X)

    def test_ordinate_init(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y)
        assert_equal(isinstance(r.ordinate, Axis), True)
        assert_equal(r.ordinate.values, TEST_Y)

    def test_ordinate_no_distribution(self):
        with self.assertRaises(ValueError):
            Relationship(utils.straight_line, TEST_X, TEST_X)

    def test_ordinate_and_ordinate_error(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_X, ordinate_error=[1]*len(TEST_X))
        assert_equal(r.x, TEST_X)
        assert_almost_equal(r.y.n, TEST_X, decimal=0)
        assert_almost_equal(r.y.ci(), np.ones((2, len(TEST_X))) * 1.96, decimal=1)

    def test_ordinate_stats(self):
        test_y = []
        for i in np.arange(1, 9, 1):
            test_y.append(scipy.stats.lognorm(i, 1, 1))
        r = Relationship(utils.straight_line, TEST_X, test_y)
        assert_equal(isinstance(r.ordinate, Axis), True)

    def test_ordinate_and_abscissa_different_length(self):
        with self.assertRaises(ValueError):
            Relationship(utils.straight_line, np.arange(1, 8, 1), TEST_Y)

    def test_ordinate_error_with_zero(self):
        with self.assertRaises(ValueError):
            Relationship(utils.straight_line, TEST_X, TEST_X, ordinate_error=np.linspace(0, 10, len(TEST_X)))

    def test_bounds_init(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10), (-1, 1)))
        assert_equal(r.bounds, ((0, 10), (-1, 1)))

    def test_variables_init(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y)
        assert_almost_equal(r.variables[0].n, 1)
        assert_almost_equal(r.variables[1].n, 1)

    def test_bounds_init_wrong_number_a(self):
        with self.assertRaises(ValueError):
            Relationship(utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10), (-1, 1), (1, 2)))

    def test_bounds_init_wrong_number_b(self):
        with self.assertRaises(ValueError):
            Relationship(utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10)))

    def test_variables_init_with_bounds(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10), (-1, 1)))
        assert_equal(np.isclose(r.variables[0].n, 5, atol=0.75), True)
        assert_equal(np.isclose(r.variables[1].n, 0, atol=0.5), True)

    def test_ln_evidence_init(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y)
        assert_equal(r.ln_evidence, None)

    def test_mcmc_results_init(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y)
        assert_equal(r.mcmc_results, None)

    def test_nested_sampling_results_init(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y)
        assert_equal(r.nested_sampling_results, None)

    def test_x(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y)
        assert_equal(isinstance(r.x, np.ndarray), True)
        assert_equal(r.x, TEST_X)

    def test_y(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y)
        assert_equal(isinstance(r.y, Axis), True)
        assert_equal(r.y.values, TEST_Y)

    def test_variable_medians(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10), (-1, 1)))
        r.max_likelihood('diff_evo')
        assert_equal(np.allclose(r.variable_medians, [1, 0], atol=1.5), True)

    def test_variable_modes(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10), (-1, 1)))
        r.max_likelihood('diff_evo')
        r.mcmc(n_burn=10, n_samples=10, progress=False, walkers=5)
        assert_equal(np.allclose(r.variable_modes, [1, 0], atol=1.5), True)

    def test_variable_modes_no_sampling(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10), (-1, 1)))
        r.max_likelihood('diff_evo')
        assert_equal(np.allclose(r.variable_modes, [1, 0], atol=1.5), True)

    def test_mcmc_done(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10), (-1, 1)))
        r.mcmc(n_burn=10, n_samples=10, progress=False, walkers=5)
        assert_equal(r.mcmc_done, True)

    def test_mcmc_not_done(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10), (-1, 1)))
        assert_equal(r.mcmc_done, False)

    def test_nested_sampling_done(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10), (-1, 1)))
        r.nested_sampling(maxiter=100, progress=False)
        assert_equal(r.nested_sampling_done, True)

    def test_nested_sampling_not_done(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10), (-1, 1)))
        assert_equal(r.nested_sampling_done, False)

    def test_get_sample(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10), (-1, 1)))
        r.nested_sampling(maxiter=100, progress=False)
        assert_equal(isinstance(r.get_sample(1), list), True)

    def test_len_parameters(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y)
        assert_equal(r.len_parameters, 2)

    def test_bayesian_information_criteria(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10), (-1, 1)))
        assert_equal(isinstance(r.bayesian_information_criteria(), float), True)

    def test_max_likelihood(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10), (-1, 1)))
        r.max_likelihood('diff_evo')
        assert_equal(isinstance(r.variables[0], Distribution), True)
        assert_equal(isinstance(r.variables[1], Distribution), True)
        assert_equal(np.isclose(r.variables[0].n, 1, atol=0.75), True)
        assert_equal(np.isclose(r.variables[1].n, 0, atol=0.75), True)

    def test_prior(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y)
        priors = r.prior()
        assert_equal(len(priors), 2)
        assert_equal(isinstance(priors[0], scipy.stats._distn_infrastructure.rv_frozen), True)
        assert_equal(isinstance(priors[1], scipy.stats._distn_infrastructure.rv_frozen), True)
        assert_equal(priors[0].interval(1), [-9, 11])
        assert_equal(priors[1].interval(1), [-9, 11])

    def test_prior_with_bounds(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10), (-1, 1)))
        priors = r.prior()
        assert_equal(len(priors), 2)
        assert_equal(isinstance(priors[0], scipy.stats._distn_infrastructure.rv_frozen), True)
        assert_equal(isinstance(priors[1], scipy.stats._distn_infrastructure.rv_frozen), True)
        assert_equal(priors[0].interval(1), [0, 10])
        assert_equal(priors[1].interval(1), [-1, 1])

    def test_mcmc(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10), (-1, 1)))
        r.mcmc(n_burn=10, n_samples=10, progress=False, walkers=5)
        assert_equal(isinstance(r.variables[0], Distribution), True)
        assert_equal(isinstance(r.variables[1], Distribution), True)
        assert_equal(r.variables[0].size, 50)
        assert_equal(r.variables[1].size, 50)
        assert_equal(r.variables[0].min > 0, True)
        assert_equal(r.variables[0].max < 10, True)
        assert_equal(r.variables[1].min > -1, True)
        assert_equal(r.variables[1].max < 1, True)

    def test_nested_sampling(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10), (-1, 1)))
        r.nested_sampling(maxiter=100, progress=False)
        assert_equal(isinstance(r.variables[0], Distribution), True)
        assert_equal(isinstance(r.variables[1], Distribution), True)
        assert_equal(r.variables[0].min > 0, True)
        assert_equal(r.variables[0].max < 10, True)
        assert_equal(r.variables[1].min > -1, True)
        assert_equal(r.variables[1].max < 1, True)

    def test_ppd(self):
        r = Relationship(utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10), (-1, 1)))
        r.mcmc(n_burn=10, n_samples=10, progress=False, walkers=5)
        ppd = r.posterior_predictive(n_posterior_samples=10, n_predictive_samples=10, progress=False)
        assert_equal(len(ppd.values), 8)
        assert_equal(ppd.values[0].size, 100)


"""
Tests for sampling module
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

import unittest
import numpy as np
from numpy.testing import assert_equal
from scipy.stats import norm
from uravu import sampling, utils
from uravu import relationship
from uravu.distribution import Distribution
import scipy.stats


TEST_Y = []
for i in np.arange(1, 9, 1):
    TEST_Y.append(Distribution(scipy.stats.norm.rvs(loc=i, scale=0.5, size=200)))
TEST_X = np.arange(1, 9, 1)

class TestSampling(unittest.TestCase):
    """
    Unit tests for optimize module.
    """

    def test_mcmc(self):
        test_rel = relationship.Relationship(
            utils.straight_line, TEST_X, TEST_Y 
        )
        test_rel.max_likelihood('mini')
        actual_results = sampling.mcmc(test_rel, n_burn=10, n_samples=10)
        assert_equal(isinstance(actual_results["distributions"][0], Distribution), True)
        assert_equal(isinstance(actual_results["distributions"][1], Distribution), True)
        assert_equal(actual_results["distributions"][0].size, 500)
        assert_equal(actual_results["distributions"][1].size, 500)

    def test_mcmc_with_other_prior(self):
        test_rel = relationship.Relationship(
            utils.straight_line, TEST_X, TEST_Y 
        )
        test_rel.max_likelihood('mini')

        def other_prior():
            """
            Another potential prior.
            """
            priors = []
            for i, variable in enumerate(test_rel.variables):
                loc = variable.n
                scale = 1
                priors.append(norm(loc=loc, scale=scale))
            return priors

        actual_results = sampling.mcmc(
            test_rel, prior_function=other_prior, n_burn=10, n_samples=10
        )
        assert_equal(isinstance(actual_results["distributions"][0], Distribution), True)
        assert_equal(isinstance(actual_results["distributions"][1], Distribution), True)
        assert_equal(actual_results["distributions"][0].size, 500)
        assert_equal(actual_results["distributions"][1].size, 500)

    def test_mcmc_with_variable_median_zero(self):
        test_rel = relationship.Relationship(
            utils.straight_line, TEST_X, TEST_Y 
        )
        test_rel.variables[0] = Distribution(np.zeros((7)))
        actual_results = sampling.mcmc(test_rel, n_burn=10, n_samples=10)
        assert_equal(isinstance(actual_results["distributions"][0], Distribution), True)
        assert_equal(isinstance(actual_results["distributions"][1], Distribution), True)
        assert_equal(actual_results["distributions"][0].size, 500)
        assert_equal(actual_results["distributions"][1].size, 500)

    def test_nested_sampling(self):
        test_rel = relationship.Relationship(
            utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10), (-1, 1)))
        actual_results = sampling.nested_sampling(test_rel, maxiter=100)
        assert_equal(isinstance(actual_results, dict), True)
        assert_equal(isinstance(actual_results["logz"][-1], float), True)
        assert_equal(isinstance(actual_results["logzerr"][-1], float), True)

    def test_nested_sampling_b_with_other_prior(self):
        test_rel = relationship.Relationship(
            utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10), (-1, 1)))
        test_rel.max_likelihood('mini')

        def other_prior():
            """
            Another potential prior.
            """
            priors = []
            for i, variable in enumerate(test_rel.variables):
                loc = variable.n
                scale = 1
                priors.append(norm(loc=loc, scale=scale))
            return priors

        actual_results = sampling.nested_sampling(
            test_rel, prior_function=other_prior, maxiter=100
        )
        assert_equal(isinstance(actual_results, dict), True)
        assert_equal(isinstance(actual_results["logz"][-1], float), True)
        assert_equal(isinstance(actual_results["logzerr"][-1], float), True)

    def test_dynamic_nested_sampling(self):
        test_rel = relationship.Relationship(
            utils.straight_line, TEST_X, TEST_Y, bounds=((0, 10), (-1, 1)))
        actual_results = sampling.nested_sampling(test_rel, dynamic=True, maxiter=100)
        assert_equal(isinstance(actual_results, dict), True)
        assert_equal(isinstance(actual_results["logz"][-1], float), True)
        assert_equal(isinstance(actual_results["logzerr"][-1], float), True)



"""
The optimize module includes the functionality necessary for maximum likelihood determination.
Furthermore, the natural log likelihood function used in the :func:`~uravu.sampling.mcmc()` and :func:`~uravu.sampling.nested_sampling()` methods may be found here.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

from typing import Callable
import numpy as np
from scipy.optimize import minimize, differential_evolution


def max_ln_likelihood(relationship: 'uravu.relationship.Relationship', method: str, x0: np.ndarray=None, **kwargs) -> np.ndarray:
    """
    Determine the variable values which maximize the likelihood for the given relationship. For keyword arguments see the :func:`scipy.optimize.minimize()` documentation.

    :param relationship: The relationship for which variables should be found.
    :param method: Method for optimisation to be used.
    :param x0: Initial guesses for the variable values. Default to the current :py:attr:`~uravu.relationship.Relationship.variables` values which are initialized as all :py:attr:`1`.

    :return: Optimized variables for the relationship.
    """
    if x0 is None:
        x0 = relationship.variable_medians
    args = (relationship.function, relationship.abscissa, relationship.ordinate)
    if method == 'diff_evo':
        res = differential_evolution(negative_lnl, relationship.bounds, args=args, **kwargs)
    elif method == 'mini':
        res = minimize(negative_lnl, x0, args=args, bounds=relationship.bounds, **kwargs)
    return res.x


def negative_lnl(variables: np.ndarray, function: Callable, abscissa: np.ndarray, ordinate: np.ndarray) -> float:
    """
    Calculate the negative natural logarithm of the likelihood given a set of variables, when there is no uncertainty in the abscissa.

    :param variables: Variables for the function evaluation.
    :param function: The function to be evaluated.
    :param abscissa: The abscissa values.
    :param ordinate: The ordinate values.

    :return: Negative natural log-likelihood between model and data.
    """
    return -ln_likelihood(variables, function, abscissa, ordinate)


def ln_likelihood(variables: np.ndarray, function: Callable, abscissa: np.ndarray, ordinate: np.ndarray) -> float:
    """
    Calculate the natural logarithm of the likelihood given a set of variables, when there is no uncertainty in the abscissa.

    :param variables: Variables for the function evaluation.
    :param function: The function to be evaluated.
    :param abscissa: The abscissa values.
    :param ordinate: The ordinate values.

    :return: Natural log-likelihood between model and data.
    """
    model = function(abscissa, *variables)
    ln_l = ordinate.logpdf(model)
    return np.sum(ln_l)



"""
The storage and manipulation of probability distributions is fundamental to the operation of ``uravu`` and Bayesian inference.
The :py:class:`~uravu.distribution.Distribution` class oversees these operations.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

from typing import Union, List
import numpy as np
from scipy.stats import normaltest, gaussian_kde
from scipy.optimize import minimize


class Distribution:
    """
    In addition to storage of the probability distribution, this class allows for some basic analysis, such as determination of normality.

    Attributes:
        samples (:py:attr:`array_like`): Samples in the distribution.
        name (:py:attr:`str`): Distribution name.
        normal (:py:attr:`bool`): Are the samples normally distributed?
        kde (:py:class:`scipy.stats.kde.gaussian_kde`): Kernel density approximation for the distribution.

    Args:
        samples (:py:attr:`array_like`): Sample for the distribution.
        name (:py:attr:`str`, optional): A name to identify the distribution. Default is :py:attr:`'Distribution'`.
        ci_points (:py:attr:`array_like`, optional): The two percentiles at which confidence intervals should be found. Default is :py:attr:`[2.5, 97.5]` (a 95 % confidence interval).

    .. _FAQ: ./faq.html
    """

    def __init__(self, samples: Union[List[Union[float, int]], np.ndarray], name: str="Distribution", random_state: np.random._generator.Generator=None) -> 'Distribution':
        """
        Initialisation function for a :py:class:`~uravu.distribution.Distribution` object.
        """
        self.name = name
        self.samples = np.array([])
        self.normal = False
        self._random_state = random_state
        if random_state is None:
            self._random_state = np.random.default_rng(np.random.randint(1))
        self.add_samples(np.array(samples))

    def to_dict(self) -> dict:
        """
        :return: Dictionary description of the object.
        """
        my_dict = {'name': self.name,
                   'samples': self.samples
                   }
        return my_dict
    
    @classmethod
    def from_dict(cls, my_dict: dict) -> 'Distribution':
        """
        :param my_dict: Dictionary description of the distribution.
        
        :return: Distribution object form the dictionary.
        """
        return cls(my_dict['samples'], name=my_dict['name'])

    @property
    def size(self) -> int:
        """
        :return: Number of samples in the distribution.
        """
        return self.samples.size

    def check_normality(self) -> bool:
        """
        Uses a :func:`scipy.stats.normaltest()` to evaluate if samples are normally distributed and updates 
        the :py:attr:`~uravu.distribution.Distribution.normal` attribute.
        """
        alpha = 0.05
        test_samples = self.samples
        if self.size > 500:
            test_samples = self._random_state.choice(self.samples, size=500)
        p_value = normaltest(test_samples)[1]
        if p_value > alpha:
            self.normal = True
        else:
            self.normal = False

    def pdf(self, x: Union[float, List[Union[float, int]], np.ndarray]) -> Union[float, np.ndarray]:
        """
        Get the probability density function for the distribution.

        :param x: Value to return probability of.
        :return: Probability.
        """
        return self.kde.pdf(x)

    def logpdf(self, x: Union[float, List[Union[float, int]], np.ndarray]) -> Union[float, np.ndarray]:
        """
        Get the natural log probability density function for the distribution.

        :param x: Value to return natural log probability of.
        :return: Natural log probability.
        """
        return self.kde.logpdf(x)

    def negative_pdf(self, x: Union[float, List[Union[float, int]], np.ndarray]) -> Union[float, np.ndarray]:
        """
        Get the negative of the probability density function for the distribution.

        :param x: Value to return negative probability of.
        :return: Negative probability.
        """
        return -self.kde.pdf(x)

    @property
    def dist_max(self) -> float:
        """
        Get the value that maximises the distribution. If no :py:attr:`kde` has been created (for example
        if the distribution has fewer than 8 values) the median is returned.

        :return: Most likely value.
        """
        try:
            return minimize(self.negative_pdf, x0=[self.n]).x[0]
        except AttributeError:
            return self.n

    @property
    def min(self) -> float:
        """
        :return: Sample minimum.
        """
        return self.samples.min()

    @property
    def max(self) -> float:
        """
        :return: Sample maximum.
        """
        return self.samples.max()

    @property
    def n(self) -> float:
        """
        :return: Median value.
        """
        return np.percentile(self.samples, [50])[0]

    @property
    def s(self) -> Union[float, None]:
        """
        :return: Standard deviation of the distribution. For a non-normal distribution, this will return :py:attr:`None`.
        """
        if self.normal:
            return np.std(self.samples, ddof=1)
        return None

    @property
    def v(self) -> Union[float, None]:
        """
        :return: Standard deviation of the distribution. For a non-normal distribution, this will return :py:attr:`None`.
        """
        if self.normal:
            return np.var(self.samples, ddof=1)
        return None

    def con_int(self, ci_points: List[float]=[2.5, 97.5]) -> np.ndarray:
        """
        Get the extrema of the confidence intervals of the distribution.

        :param ci_points: The confidence interval points to return.
        :return: Distribution values at the confidence interval.
        """
        return self.ci(ci_points)

    def ci(self, ci_points: List[float]=[2.5, 97.5]) -> np.ndarray:
        """
        Get the extrema of the confidence intervals of the distribution.

        :param ci_points: The confidence interval points to return.
        :return: Distribution values at the confidence interval.
        """
        return np.percentile(self.samples, ci_points)

    def add_samples(self, samples: Union[List[Union[float, int]], np.ndarray]):
        """
        Add samples to the distribution.

        Args:
            samples (:py:attr:`array_like`): Samples to be added to the distribution.
        """
        self.samples = np.append(self.samples, np.array(samples).flatten())
        if (self.size > 8) and (np.std(self.samples) != 0):
            self.check_normality()
            self.kde = gaussian_kde(self.samples)

    def __repr__(self) -> np.ndarray:
        """
        :return: Representation.
        """
        return self.samples.__repr__()

    def __array__(self) -> np.ndarray:
        """
        :return: Array of samples.
        """
        return self.samples

    def __add__(self, obj2: Union[float, np.ndarray]) -> np.ndarray:
        """
        :return: Sum of samples and obj2.
        """
        return np.add(self, obj2)
 
    def __sub__(self, obj2: Union[float, np.ndarray]) -> np.ndarray:
        """
        :return: Difference between samples and obj2.
        """
        return np.subtract(self, obj2)
 
    def __mul__(self, obj2: Union[float, np.ndarray]) -> np.ndarray:
        """
        :return: Product of samples and obj2.
        """
        return np.multiply(self, obj2)
 
    def __truediv__(self, obj2: Union[float, np.ndarray]) -> np.ndarray:
        """
        :return: Ratio between samples and obj2.
        """
        return np.divide(self, obj2)
   
    def __mod__(self, obj2: Union[float, np.ndarray]) -> np.ndarray:
        """
        :return: Modulus of samples by obj2.
        """
        return np.remainder(self, obj2)

    def __str__(self) -> np.ndarray:
        """
        :return: String representation.
        """
        if self.s is not None:
            return f'{self.n:.3e} +/- {self.s:.3e}'
        else:
            c = np.abs(self.n - self.con_int())
            return f'{self.n:.3e}(+{c[1]:.3e}/-{c[0]:.3e}'



"""
Tests for utils module
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

import unittest
import numpy as np
import scipy.stats
from uncertainties import ufloat
from numpy.testing import assert_almost_equal, assert_equal
from uravu import utils
from uravu.relationship import Relationship
from uravu.distribution import Distribution


class TestUtils(unittest.TestCase):
    """
    Tests for the relationship module and class.
    """

    def test_straight_line(self):
        """
        Test straight line function.
        """
        test_x = np.linspace(0, 9, 10)
        test_y = np.linspace(0, 18, 10)
        result_y = utils.straight_line(test_x, 2, 0)
        assert_almost_equal(result_y, test_y)

    def test_bayes_factor_a(self):
        """
        Test the bayes factor function.
        """
        model_1 = 2
        model_2 = 4
        expected_result = -4
        actual_result = utils.bayes_factor(model_1, model_2)
        assert_almost_equal(actual_result, expected_result)

    def test_bayes_factor_b(self):
        """
        Test the bayes factor function with uncertainties.
        """
        model_1 = ufloat(2, 1)
        model_2 = ufloat(4, 1)
        expected_result = ufloat(-4, 2.82842712474619032)
        actual_result = utils.bayes_factor(model_1, model_2)
        assert_almost_equal(actual_result.n, expected_result.n)
        assert_almost_equal(actual_result.s, expected_result.s)

    def test_correlation_matrix(self):
        """
        Test correlation_matrix function.
        """
        TEST_Y = []
        for i in np.arange(1, 9, 1):
            TEST_Y.append(Distribution(scipy.stats.norm.rvs(loc=i, scale=0.5, size=200)))
        TEST_X = np.arange(1, 9, 1)
        test_rel = Relationship(utils.straight_line, TEST_X, TEST_Y)
        test_rel.max_likelihood('mini')
        test_rel.mcmc(n_burn=10, n_samples=10)
        actual_matrix = utils.correlation_matrix(test_rel)
        assert_equal(actual_matrix.shape, (2, 2))
        assert_almost_equal(actual_matrix[1, 0], actual_matrix[0, 1])
        assert_almost_equal(actual_matrix[0, 0], 1.0)
        assert_almost_equal(actual_matrix[1, 1], 1.0)
        assert_equal(test_rel.mcmc_done, True)

    def test_normal_latex(self):
        """
        Test the latex print from a normal distribution
        """
        np.random.seed(1)
        d = Distribution(scipy.stats.norm.rvs(loc=0, scale=0.5, size=2000))
        assert_equal(utils.latex(d), r'$1.725e-02\pm{9.665e-01}$')

    def test_not_normal_latex(self):
        """
        Test the latex print from a non-normal distribution
        """
        np.random.seed(1)
        d = Distribution(
            scipy.stats.lognorm.rvs(loc=10, scale=0.5, s=10, size=2000))
        assert_equal(utils.latex(d), r'$1.071e+01^{+1.754e+08}_{-7.060e-01}$')



"""
The :py:class:`~uravu.axis.Axis` class controls the organisation of axes in the uravu code, including the evaluation of an axis-level multidimensional kernal density estimate.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey


from typing import Union, List, Tuple
import numpy as np
from scipy.stats import gaussian_kde
from uravu.distribution import Distribution


class Axis:
    """
    The Axes class is a flexible storage option for both numerical (:py:class:`numpy.ndarray`) and distribution (:py:class:`uravu.distribution.Distribution`) arrays.

    :param values: Array of values.
    """
    def __init__(self, values: Union[list, Distribution]) -> 'Axis':
        """
        Initialisation function for a :py:class:`~uravu.axis.Axis` object.
        """
        self._kde = None
        if isinstance(values[0], Distribution):
            self._values = values
            self._kde = _get_kde(self.values)
        else:
            self._values = np.array(values)

    def to_dict(self) -> dict:
        """
        :return: Dictionary of Axis.
        """
        if isinstance(self.values[0], Distribution):
            return {'values': [i.to_dict() for i in self.values]}
        else: 
            return {'values': self.values.tolist()}

    @classmethod
    def from_dict(cls, my_dict: dict) -> 'Axis':
        """
        Class method to produce from a dictionary.

        :param my_dict: Input dictionary
        
        :return: Axis from dictionary.
        """
        if isinstance(my_dict['values'][0], dict):
            v = [Distribution.from_dict(i) for i in my_dict['values']]
            return Axis(v)
        else:
            return Axis(my_dict['values'])

    @property
    def kde(self) -> 'scipy.stats._kde.gaussian_kde':
        """
        :return: Multi-dimensional kernel density estimation for the axis.
        """
        return self._kde

    @property
    def values(self) -> Union[List[Distribution], np.ndarray]:
        """
        :return: Array of values.
        """
        return self._values

    @property
    def n(self) -> np.ndarray:
        """
        :return: Medians for axis.
        """
        v = np.zeros(self.shape)
        if isinstance(self.values[0], Distribution):
            for i, o in enumerate(self.values):
                v[i] = o.n
            return v
        return self.values

    @property
    def s(self) -> np.ndarray:
        """
        :return: Standard deviation of axis values.
        """
        if isinstance(self.values[0], Distribution):
            dv = np.zeros((self.size))
            for i, o in enumerate(self.values):
                if o.s is None:
                    dv[i] = np.nan
                else:
                    dv[i] = o.s
            return dv
        return np.zeros(self.shape)

    def ci(self, ci_points: List[float]=[2.5, 97.5]) -> np.ndarray:
        """
        :return: Uncertainties from 95 % confidence intervals for axis.
        """
        if isinstance(self.values[0], Distribution):
            dv = np.zeros((2, self.size))
            for i, o in enumerate(self.values):
                dv[:, i] = np.abs(o.con_int(ci_points) - o.n)
            return dv
        return np.zeros(self.shape)

    @property
    def mode(self) -> np.ndarray:
        """
        :return: Values that maximise the probability for axis.
        """
        v = np.zeros(self.shape)
        if isinstance(self.values[0], Distribution):
            for i, o in enumerate(self.values):
                v[i] = o.dist_max
            return v
        return self.values

    @property
    def size(self) -> int:
        """
        :return: Size of axis.
        """
        if isinstance(self.values[0], Distribution):
            return len(self.values)
        return self.values.size

    @property
    def shape(self) -> Union[int, Tuple[int]]:
        """
        :return: Shape of axis.
        """
        if isinstance(self.values[0], Distribution):
            return len(self.values)
        return self.values.shape

    def pdf(self, x: np.ndarray) -> np.ndarray:
        """"
        Get the probability density function for all of the distributions in the axes.

        :param x: Values to return probability of.

        :return: Probability.
        """
        return self.kde.pdf(x)

    def logpdf(self, x: np.ndarray) -> np.ndarray:
        """"
        Get the natural log probability density function for all of the distributions in the axes.

        :param x: Values to return natural log probability of.

        :return: Natural log probability.
        """
        return self.kde.logpdf(x)


def _get_kde(values: np.ndarray) -> 'scipy.stats._kde.gaussian_kde':
    """
    Determine the kernel density estimate for a given set of values.

    :param values: Sample for kde.

    :return: Kernel density estimate for samples.
    """
    min_size = values[0].size
    for v in values:
        if v.size < min_size:
            min_size = v.size
    random_list = [np.random.choice(v.samples, size=min_size) for v in values]
    return gaussian_kde(np.vstack(random_list))



"""
Tests for axis module
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey


import unittest
import numpy as np
from numpy.testing import assert_almost_equal, assert_equal
from uravu.distribution import Distribution
import scipy.stats
from uravu.axis import Axis
from scipy.stats import norm, uniform, gaussian_kde


DISTRO1 = Distribution(norm.rvs(loc=0, scale=1, size=10000, random_state=np.random.RandomState(1)))
DISTRO2 = Distribution(norm.rvs(loc=1, scale=1, size=10000, random_state=np.random.RandomState(2)))
AX = Axis([DISTRO1, DISTRO2])

AX_ARRAY = Axis([0, 1])


class TestAxis(unittest.TestCase):
    """
    Testing the Axis class.
    """
    def test_init_values(self):
        assert_equal(AX.values[0].samples, DISTRO1.samples)
        assert_equal(AX.values[1].samples, DISTRO2.samples)

    def test_dictionary_roundtrip(self):
        AX2 = Axis.from_dict(AX.to_dict())
        assert_equal(AX2.values[0].samples, AX.values[0].samples)
        assert_equal(AX2.values[1].samples, AX.values[1].samples)
        assert_equal(AX2.kde.neff, AX.kde.neff)

    def test_dictionary_roundtrip_array(self):
        AX2 = Axis.from_dict(AX_ARRAY.to_dict())
        assert_equal(AX2.values[0], AX_ARRAY.values[0])
        assert_equal(AX2.values[1], AX_ARRAY.values[1])
        assert AX_ARRAY.kde is None
        assert AX2.kde is None
    
    def test_init_kde(self):
        assert_equal(isinstance(AX.kde, gaussian_kde), True)

    def test_init_kde_size_change(self):
        distro2 = Distribution(norm.rvs(loc=1, scale=1, size=1000, random_state=np.random.RandomState(2)))
        AX = Axis([DISTRO1, distro2])
        assert_equal(AX.values[1].samples, distro2.samples)

    def test_n(self):
        assert_almost_equal(AX.n, [0, 1], decimal=1)

    def test_n_array(self):
        assert_equal(AX_ARRAY.n, [0, 1])

    def test_s(self):
        assert_almost_equal(AX.ci(), [[1.96, 1.96], [1.96, 1.96]], decimal=1)

    def test_s_array(self):
        assert_equal(AX_ARRAY.ci(), [0, 0])

    def test_mode(self):
        assert_almost_equal(AX.mode, [0, 1], decimal=1)

    def test_mode_array(self):
        assert_equal(AX_ARRAY.mode, [0, 1])

    def test_size(self):
        assert_equal(AX.size, 2)

    def test_size_array(self):
        assert_equal(AX_ARRAY.size, 2)

    def test_shape(self):
        assert_equal(AX.shape, 2)

    def test_shape_array(self):
        ax = Axis(np.ones((3, 3)))
        assert_equal(ax.shape, (3, 3))

    def test_pdf(self):
        assert_almost_equal(AX.pdf([0, 1]), [0.1495], decimal=0)

    def test_logpdf(self):
        assert_almost_equal(AX.logpdf([0, 1]), np.log([0.1495]), decimal=1)
